"""Models for MyLightSystems API Client."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import datetime


@dataclass
class Auth:
    """Auth model."""

    token: str


@dataclass
class Profile:
    """Profile model."""

    id: int
    grid_type: str


@dataclass
class Device:
    """Device model."""

    id: str
    name: str
    type: str
    type_id: str


@dataclass
class BatteryDevice(Device):
    """Represent a battery."""

    state: bool
    capacity: int


@dataclass
class RelayDevice(Device):
    """Represent a relay."""

    state: bool
    master_id: str
    master_type: str


@dataclass
class CounterDevice(Device):
    """Represent a counter."""

    state: bool
    phase: int
    master_id: str
    master_type: str


@dataclass
class CompositeCounterDevice(Device):
    """Represent a composite counter."""

    master_id: str
    master_type: str
    children: dict[str, int]


@dataclass
class VirtualDevice(Device):
    """Represent a virtual."""

    state: bool


@dataclass
class MasterDevice(Device):
    """Represent a master."""

    state: bool
    report_period: int


@dataclass
class Measure:
    """Represent a measure."""

    type: str
    value: float
    unit: str


@dataclass
class SensorMeasure:
    """Represent a sensor measure."""

    type: str | None
    value: float
    unit: str | None
    date: datetime


@dataclass
class SensorState:
    """Represent a sensor state."""

    sensor_id: str
    measure: SensorMeasure


@dataclass
class DeviceState:
    """Represent a device state."""

    device_id: str
    report_period: int
    state: bool
    sensor_states: list[SensorState]


@dataclass
class SwitchState:
    """Represent the state of the switch."""

    state: bool



"""Client for MyLightSystems API."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
import logging
import socket
from typing import TYPE_CHECKING, Any

from aiohttp import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET
from yarl import URL

from mylightsystems.const import (
    AUTH_URL,
    DEFAULT_BASE_URL,
    DEVICES_URL,
    MEASURES_TOTAL_URL,
    PROFILE_URL,
    STATES_URL,
    SWITCH_URL,
)
from mylightsystems.device_factory import DeviceFactory
from mylightsystems.exceptions import (
    MyLightSystemsConnectionError,
    MyLightSystemsInvalidAuthError,
    MyLightSystemsMeasuresTotalNotSupportedError,
    MyLightSystemsSwitchNotAllowedError,
    MyLightSystemsUnauthorizedError,
    MyLightSystemsUnknownDeviceError,
)
from mylightsystems.models import (
    Auth,
    Device,
    DeviceState,
    Measure,
    Profile,
    SensorMeasure,
    SensorState,
    SwitchState,
)

if TYPE_CHECKING:
    from typing import Self

_LOGGER = logging.getLogger(__name__)


@dataclass
class MyLightSystemsApiClient:
    """Main class for handling communication with MyLightSystems API."""

    base_url: str = DEFAULT_BASE_URL
    session: ClientSession | None = None
    request_timeout: int = 10
    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to the MyLightSystems API."""
        url = URL(self.base_url).with_path(uri)

        headers = {
            "Content-Type": "application/json",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method, url, headers=headers, params=params
                )

                _LOGGER.debug(
                    "Data retrieved from %s, status: %s", url, response.status
                )

                response.raise_for_status()

                json_response = await response.json()

                if (
                    json_response["status"] == "error"
                    and json_response["error"] == "not.authorized"
                ):
                    raise MyLightSystemsUnauthorizedError

                return json_response

        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to the device"
            raise MyLightSystemsConnectionError(msg) from exception
        except (
            ClientError,
            ClientResponseError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with the device"
            raise MyLightSystemsConnectionError(msg) from exception

    async def auth(self, email: str, password: str) -> Auth:
        """Login to MyLightSystems API."""
        response = await self._request(
            AUTH_URL,
            params={"email": email, "password": password},
        )

        if response["status"] == "error" and response["error"] in (
            "invalid.credentials",
            "undefined.email",
            "undefined.password",
        ):
            raise MyLightSystemsInvalidAuthError

        return Auth(token=response["authToken"])

    async def get_profile(self, auth_token: str) -> Profile:
        """Get user profile."""
        response = await self._request(
            PROFILE_URL,
            params={"authToken": auth_token},
        )

        return Profile(id=response["id"], grid_type=response["gridType"])

    async def get_devices(self, auth_token: str) -> list[Device]:
        """Get devices."""
        response = await self._request(
            DEVICES_URL,
            params={"authToken": auth_token},
        )

        device_factory = DeviceFactory()

        return [
            device_factory.create_device(data=device) for device in response["devices"]
        ]

    async def get_measures_total(
        self, auth_token: str, device_id: str
    ) -> list[Measure]:
        """Get measures total."""
        response = await self._request(
            MEASURES_TOTAL_URL,
            params={"authToken": auth_token, "device_id": device_id},
        )

        if (
            response["status"] == "error"
            and response["error"] == "device.not.supports.total.measures"
        ):
            raise MyLightSystemsMeasuresTotalNotSupportedError

        return [
            Measure(type=measure["type"], unit=measure["unit"], value=measure["value"])
            for measure in response["measure"]["values"]
        ]

    async def get_states(self, auth_token: str) -> list[DeviceState]:
        """Get states."""
        response = await self._request(
            STATES_URL,
            params={"authToken": auth_token},
        )

        return [
            DeviceState(
                device_id=device_state["deviceId"],
                report_period=device_state["effectiveReportPeriod"],
                sensor_states=[
                    SensorState(
                        sensor_id=sensor_state["sensorId"],
                        measure=SensorMeasure(
                            value=sensor_state["measure"]["value"],
                            type=sensor_state["measure"].get("type", None),
                            unit=sensor_state["measure"].get("unit", None),
                            date=sensor_state["measure"]["date"],
                        ),
                    )
                    for sensor_state in device_state["sensorStates"]
                ],
                state=device_state["state"].lower() == "on",
            )
            for device_state in response["deviceStates"]
        ]

    async def switch(self, auth_token: str, device_id: str, value: bool) -> SwitchState:  # noqa: FBT001
        """Change switch state."""
        response = await self._request(
            SWITCH_URL,
            params={"authToken": auth_token, "id": device_id, "on": str(value).lower()},
        )

        if response["status"] == "error":
            if response["error"] == "switch.not.allowed":
                raise MyLightSystemsSwitchNotAllowedError
            if response["error"] == "device.not.found":
                raise MyLightSystemsUnknownDeviceError

        return SwitchState(state=response["state"] == "on")

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The AirGradientClient object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Exceptions for MyLightSystems API Client."""


class MyLightSystemsError(Exception):
    """Generic exception."""


class MyLightSystemsConnectionError(MyLightSystemsError):
    """Connection error."""


class MyLightSystemsInvalidAuthError(MyLightSystemsError):
    """Invalid authentication error."""


class MyLightSystemsUnauthorizedError(MyLightSystemsError):
    """Unauthorized error."""


class MyLightSystemsUnknownDeviceError(MyLightSystemsError):
    """Unknown device error."""


class MyLightSystemsMeasuresTotalNotSupportedError(MyLightSystemsError):
    """Device doesn't support measures total error."""


class MyLightSystemsSwitchNotAllowedError(MyLightSystemsError):
    """Switch not allowed error."""



"""Constants for MyLightSystems Api Client."""

DEFAULT_TIMEOUT_IN_SECONDS: int = 10

DEFAULT_BASE_URL: str = "https://myhome.mylight-systems.com"
AUTH_URL: str = "/api/auth"
PROFILE_URL: str = "/api/profile"
DEVICES_URL: str = "/api/devices"
MEASURES_TOTAL_URL: str = "/api/measures/total"
STATES_URL: str = "/api/states"
SWITCH_URL: str = "/api/device/switch"



"""Asynchronous Python client for MyLightSystems API."""

from mylightsystems.client import MyLightSystemsApiClient
from mylightsystems.exceptions import (
    MyLightSystemsConnectionError,
    MyLightSystemsError,
    MyLightSystemsInvalidAuthError,
    MyLightSystemsUnauthorizedError,
)
from mylightsystems.models import Auth

__all__ = [
    "Auth",
    "MyLightSystemsApiClient",
    "MyLightSystemsConnectionError",
    "MyLightSystemsError",
    "MyLightSystemsInvalidAuthError",
    "MyLightSystemsUnauthorizedError",
]



"""Device factory for MyLightSystems API."""

from __future__ import annotations

import dataclasses
from typing import Any

from mylightsystems.exceptions import MyLightSystemsUnknownDeviceError
from mylightsystems.models import (
    BatteryDevice,
    CompositeCounterDevice,
    CounterDevice,
    Device,
    MasterDevice,
    RelayDevice,
    VirtualDevice,
)

# Mapping from type name to class
device_type_mapping: dict[str, type[Device]] = {
    "bat": BatteryDevice,
    "sw": RelayDevice,
    "cmp": CounterDevice,
    "gmd": CompositeCounterDevice,
    "vrt": VirtualDevice,
    "mst": MasterDevice,
}

device_property_mapping: dict[str, str] = {
    "id": "id",
    "name": "name",
    "type": "type",
    "state": "state",
}


class DeviceFactory:  # pylint: disable=too-few-public-methods
    """Device factory."""

    def __init__(self) -> None:
        """Device factory initializer."""
        self.type_mapping: dict[str, type[Device]] = {
            "bat": BatteryDevice,
            "sw": RelayDevice,
            "cmp": CounterDevice,
            "gmd": CompositeCounterDevice,
            "vrt": VirtualDevice,
            "mst": MasterDevice,
        }
        self.field_mapping = {
            "deviceTypeId": "type_id",
            "masterMac": "master_id",
            "masterType": "master_type",
            "reportPeriod": "report_period",
            "batteryCapacity": "capacity",
        }
        self.value_transformers: dict[str, Any] = {
            "state": lambda x: x.lower() == "on",
            "children": lambda x: {item["mac"]: item["phase"] for item in x},
        }

    def create_device(self, data: dict[str, Any]) -> Device:
        """Create a new device."""
        device_type = data.get("type", "device").lower()
        device_class = self.type_mapping.get(device_type, None)

        if device_class is None:
            raise MyLightSystemsUnknownDeviceError(device_type)

        # Get the fields of the device class
        class_fields = {field.name for field in dataclasses.fields(device_class)}

        # Create a dictionary with only the relevant fields
        device_data = {}
        for json_field, value in data.items():
            class_field = self.field_mapping.get(json_field, json_field)
            if class_field in class_fields:
                # Apply value transformation if needed
                if class_field in self.value_transformers:
                    value = self.value_transformers[class_field](value)  # noqa: PLW2901
                device_data[class_field] = value

        return device_class(**device_data)
